@@ -17,6 +17,7 @@
 
 #include "chan_h323.h"
 
+#define AST_MODULE "chan_ooh323"
 
 /* Defaults */
 #define DEFAULT_CONTEXT "default"
@@ -35,7 +36,6 @@
 #define H323_DISABLEGK          (1<<7)
 
 /* Channel description */
-static const char desc[] = "Objective Systems H323 Channel";
 static const char type[] = "OOH323";
 static const char tdesc[] = "Objective Systems H323 Channel Driver";
 static const char config[] = "ooh323.conf";
@@ -44,21 +44,24 @@ static const char config[] = "ooh323.conf";
 /* Channel Definition */
 static struct ast_channel *ooh323_request(const char *type, int format, 
                                         void *data, int *cause);
-static int ooh323_digit(struct ast_channel *ast, char digit);
+static int ooh323_digit_begin(struct ast_channel *ast, char digit);
+static int ooh323_digit_end(struct ast_channel *ast, char digit, unsigned int duration);
 static int ooh323_call(struct ast_channel *ast, char *dest, int timeout);
 static int ooh323_hangup(struct ast_channel *ast);
 static int ooh323_answer(struct ast_channel *ast);
 static struct ast_frame *ooh323_read(struct ast_channel *ast);
 static int ooh323_write(struct ast_channel *ast, struct ast_frame *f);
-static int ooh323_indicate(struct ast_channel *ast, int condition);
+static int ooh323_indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen);
 static int ooh323_fixup
    (struct ast_channel *oldchan, struct ast_channel *newchan);
 
-static struct ast_rtp *ooh323_get_rtp_peer(struct ast_channel *chan);
-static struct ast_rtp *ooh323_get_vrtp_peer(struct ast_channel *chan);
+static enum ast_rtp_get_result ooh323_get_rtp_peer(struct ast_channel *chan, struct ast_rtp **rtp);
+static enum ast_rtp_get_result ooh323_get_vrtp_peer(struct ast_channel *chan, struct ast_rtp **rtp);
 static int ooh323_set_rtp_peer(struct ast_channel *chan, struct ast_rtp *rtp, 
                              struct ast_rtp *vrtp, int codecs, int nat_active);
 
+static void print_codec_to_cli(int fd, struct ast_codec_pref *pref);
+static void ast_ooh323c_exit();
 
 
 static const struct ast_channel_tech ooh323_tech = {
@@ -66,7 +69,8 @@ static const struct ast_channel_tech ooh323_tech = {
    .description = tdesc,
    .capabilities = -1,
    .requester = ooh323_request,
-   .send_digit = ooh323_digit,
+   .send_digit_begin = ooh323_digit_begin,
+   .send_digit_end = ooh323_digit_end,
    .call = ooh323_call,
    .hangup = ooh323_hangup,
    .answer = ooh323_answer,
@@ -80,10 +84,10 @@ static const struct ast_channel_tech ooh323_tech = {
 };
 
 static struct ast_rtp_protocol ooh323_rtp = {
-   .type = type,
-   .get_rtp_info = ooh323_get_rtp_peer,
-   .get_vrtp_info = ooh323_get_vrtp_peer,
-   .set_rtp_peer=  ooh323_set_rtp_peer
+   type: type,
+   get_rtp_info: ooh323_get_rtp_peer,
+   get_vrtp_info: ooh323_get_vrtp_peer,
+   set_rtp_peer: ooh323_set_rtp_peer
 };
 
 /* H.323 channel private structure */
@@ -141,6 +145,8 @@ struct ooh323_user{
    struct ast_codec_pref prefs;
    int         dtmfmode;
    int         rtptimeout;
+   int         mUseIP;        /* Use IP address or H323-ID to search user */
+   char        mIP[20];
    struct ooh323_user *next;
 };
 
@@ -155,6 +161,7 @@ struct ooh323_peer{
    char        accountcode[20];
    int         amaflags;
    int         dtmfmode;
+   int         mFriend;    /* indicates defined as friend */
    char        ip[20];
    int         port;
    char        *h323id;    /* H323-ID alias, which asterisk will register with gk to reach this peer*/
@@ -207,6 +214,7 @@ static enum RasGatekeeperMode gRasGkMode = RasNoGatekeeper;
 static int  gIsGateway = 0;
 static int  gFastStart = 1;
 static int  gTunneling = 1;
+static int  gMediaWaitForConnect = 0;
 static int  gTOS = 0;
 static int  gRTPTimeout = 60;
 static char gAccountcode[80] = DEFAULT_H323ACCNT;
@@ -242,22 +250,19 @@ static struct ast_channel *ooh323_new(struct ooh323_pvt *i, int state,
 
    /* Don't hold a h323 pvt lock while we allocate a channel */
    ast_mutex_unlock(&i->lock);
-   ch = ast_channel_alloc(1);
+   ch = ast_channel_alloc(1, state, i->callerid_num, i->callerid_name, i->accountcode, i->exten, i->context, i->amaflags, "OOH323/%s-%04x", host, rand() & 0xffff);
    ast_mutex_lock(&i->lock);
 
    if(ch)
    {
       ast_mutex_lock(&ch->lock);
       ch->tech = &ooh323_tech;
-      snprintf(ch->name, sizeof(ch->name), "OOH323/%s-%04x", 
-                                                     host, rand() & 0xffff);
 
       ch->nativeformats = i->capability;
 
 
       fmt = ast_best_codec(ch->nativeformats);
 
-      ch->type = type;
       ch->fds[0] = ast_rtp_fd(i->rtp);
       ch->fds[1] = ast_rtcp_fd(i->rtp);
 
@@ -320,10 +325,9 @@ static struct ast_channel *ooh323_new(struct ooh323_pvt *i, int state,
          }
       }
 
-      if (!ast_strlen_zero(i->accountcode)){
-         strncpy(ch->accountcode, i->accountcode, sizeof(ch->accountcode)-1);
-         ch->accountcode[sizeof(ch->accountcode)-1] = '\0';
-      }
+      if (!ast_strlen_zero(i->accountcode))
+         ast_string_field_set(ch, accountcode, i->accountcode);
+      
       if (i->amaflags)
          ch->amaflags = i->amaflags;
 
@@ -459,6 +463,9 @@ static struct ast_channel *ooh323_request(const char *type, int format,
    char tmp[256];
    char formats[512];
    int oldformat;
+   char *sport = NULL;
+   int port = 0;
+
    if(gH323Debug)
       ast_verbose("---   ooh323_request - data %s format %s\n", (char*)data,  
                               ast_getformatname_multiple(formats,512,format) );
@@ -503,9 +510,16 @@ static struct ast_channel *ooh323_request(const char *type, int format,
       dest = tmp;
       ext = NULL;
    }
+   
+   /*if((sport = strchr(dest, ':'))) {
+      *sport = '\0';
+      sport++;
+      port = atoi(sport);
+   }*/
 
-   if(dest)
-      peer = find_peer(dest);
+   if(dest) {
+      peer = find_peer(dest, port);
+   }
    else{
       ast_log(LOG_ERROR, "Destination format is not supported\n");
       return NULL;
@@ -525,7 +539,7 @@ static struct ast_channel *ooh323_request(const char *type, int format,
       if(peer->capability & format)
          p->capability = peer->capability & format;
       else{
-        p->capability = peer->capability;
+         p->capability = peer->capability;
       }
       memcpy(&p->prefs, &peer->prefs, sizeof(struct ast_codec_pref));
       p->dtmfmode = peer->dtmfmode;
@@ -538,11 +552,14 @@ static struct ast_channel *ooh323_request(const char *type, int format,
 
       memcpy(&p->prefs, &gPrefs, sizeof(struct ast_codec_pref));
       p->username = strdup(dest);
-
      
       p->host = strdup(dest);
-      if(ext)
+      if(port > 0) {
+         p->port = port;
+      }
+      if(ext) {
          strncpy(p->exten, ext, sizeof(p->exten)-1);
+      }
    }
 
 
@@ -593,64 +610,115 @@ static struct ooh323_pvt* find_call(ooCallData *call)
    return p;
 }
 
-struct ooh323_user *find_user(const char * name)
+struct ooh323_user *find_user(const char * name, const char* ip)
 {
    struct ooh323_user *user=NULL;
 
    if(gH323Debug)
-      ast_verbose("---   find_user\n");
-
+      ast_verbose("---   find_user \n");
 
    user = userl.users;
    ast_mutex_lock(&userl.lock);
    while(user)
    {
-      if(name && !strcmp(user->name, name))
+      if(ip && user->mUseIP && !strcmp(user->mIP, ip)) {
          break;
+      }
+      if(name && !strcmp(user->name, name)) {
+         break;
+      }
       user = user->next;
    }
    ast_mutex_unlock(&userl.lock);
+
    if(gH323Debug)
       ast_verbose("+++   find_user\n");
 
    return user;      
 }
 
-struct ooh323_peer *find_peer(const char * name)
+struct ooh323_peer *find_friend(const char *name, int port)
 {
-   struct ooh323_peer *peer=NULL;
+   struct ooh323_peer *peer=NULL;  
 
    if(gH323Debug)
-      ast_verbose("---   find_peer\n");
+      ast_verbose("---   find_friend \"%s\"\n", name);
 
 
    peer = peerl.peers;
    ast_mutex_lock(&peerl.lock);
    while(peer)
    {
+      if(gH323Debug) {
+         ast_verbose("      comparing with \"%s\"\n", peer->ip);
+      }
+      if(!strcmp(peer->ip, name)) {
+         if(port > 0 && peer->port == port) { break; }
+         else if (port <= 0) { break; }
+      }
+      peer = peer->next;
+   }
+   ast_mutex_unlock(&peerl.lock);
+
+   if(gH323Debug) {
+      if(peer) {
+         ast_verbose("      found matching friend\n");
+      }
+      ast_verbose("+++   find_friend \"%s\"\n", name);
+   }
+
+   return peer;      
+}
+
+
+struct ooh323_peer *find_peer(const char * name, int port)
+{
+   struct ooh323_peer *peer=NULL;  
+
+   if(gH323Debug)
+      ast_verbose("---   find_peer \"%s\"\n", name);
+
+
+   peer = peerl.peers;
+   ast_mutex_lock(&peerl.lock);
+   while(peer)
+   {
+      if(gH323Debug) {
+         ast_verbose("      comparing with \"%s\"\n", peer->ip);
+      }
       if(!strcasecmp(peer->name, name))
          break;
       if(peer->h323id && !strcasecmp(peer->h323id, name))
          break;
       if(peer->e164 && !strcasecmp(peer->e164, name))
          break;
+      /*
+      if(!strcmp(peer->ip, name)) {
+         if(port > 0 && peer->port == port) { break; }
+         else if (port <= 0) { break; }
+      }
+      */
       peer = peer->next;
    }
    ast_mutex_unlock(&peerl.lock);
 
-   if(gH323Debug)
-      ast_verbose("+++   find_peer\n");
+   if(gH323Debug) {
+      if(peer) {
+         ast_verbose("      found matching peer\n");
+      }
+      ast_verbose("+++   find_peer \"%s\"\n", name);
+   }
 
    return peer;      
 }
 
-static int ooh323_digit(struct ast_channel *chan, char digit)
+static int ooh323_digit_begin(struct ast_channel *chan, char digit)
 {
    char dtmf[2];
    struct ooh323_pvt *p = (struct ooh323_pvt *) chan->tech_pvt;
    
    if(gH323Debug)
-      ast_verbose("---   ooh323_digit\n");
+      ast_verbose("---   ooh323_digit_begin\n");
 
    if(!p){
       ast_log(LOG_ERROR, "No private structure for call\n");
@@ -658,7 +726,7 @@ static int ooh323_digit(struct ast_channel *chan, char digit)
    }
    ast_mutex_lock(&p->lock);
    if (p->rtp && (p->dtmfmode & H323_DTMF_RFC2833)) {
-      ast_rtp_senddigit(p->rtp, digit);
+      ast_rtp_senddigit_begin(p->rtp, digit);
    }
    else if (((p->dtmfmode & H323_DTMF_Q931) ||
                    (p->dtmfmode & H323_DTMF_H245ALPHANUMERIC) ||
@@ -671,7 +739,30 @@ static int ooh323_digit(struct ast_channel *chan, char digit)
    }
    ast_mutex_unlock(&p->lock);
    if(gH323Debug)
-      ast_verbose("+++   ooh323_digit\n");
+      ast_verbose("+++   ooh323_digit_begin\n");
+
+   return 0;
+}
+
+static int ooh323_digit_end(struct ast_channel *chan, char digit, unsigned int duration)
+{
+   char dtmf[2];
+   struct ooh323_pvt *p = (struct ooh323_pvt *) chan->tech_pvt;
+
+   if(gH323Debug)
+      ast_verbose("---   ooh323_digit_end\n");
+
+   if(!p){
+      ast_log(LOG_ERROR, "No private structure for call\n");
+      return -1;
+   }
+   ast_mutex_lock(&p->lock);
+   if (p->rtp && (p->dtmfmode & H323_DTMF_RFC2833)) 
+      ast_rtp_senddigit_end(p->rtp, digit);
+
+   ast_mutex_unlock(&p->lock);
+   if(gH323Debug)
+      ast_verbose("+++   ooh323_digit_end\n");
 
    return 0;
 }
@@ -682,7 +773,7 @@ static int ooh323_call(struct ast_channel *ast, char *dest, int timeout)
    struct ooh323_pvt *p = ast->tech_pvt;
    char destination[256];
    int res=0;
-   char *val=NULL;
+   const char *val=NULL;
    ooCallOptions opts = {
       .fastStart = TRUE,
       .tunneling = TRUE,
@@ -788,15 +879,15 @@ static int ooh323_hangup(struct ast_channel *ast)
       ast_mutex_lock(&p->lock);
 
       if(gH323Debug)
-         ast_verbose("    hanging %s\n", p->username);
-
+	ast_verbose("    hanging %s\n", p->username);
+      ast->tech_pvt = NULL; 
       if(!ast_test_flag(p, H323_ALREADYGONE)) {
          ast_mutex_lock(&ooh323c_cmd_lock);
          ooHangCall(p->callToken, 
              ooh323_convert_hangupcause_asteriskToH323(p->owner->hangupcause));
          ast_mutex_unlock(&ooh323c_cmd_lock);
          ast_set_flag(p, H323_ALREADYGONE);
-         //ast_mutex_unlock(&p->lock);
+         /* ast_mutex_unlock(&p->lock); */
       }
       else {
          ast_set_flag(p, H323_NEEDDESTROY);
@@ -817,11 +908,13 @@ static int ooh323_hangup(struct ast_channel *ast)
       ast_update_use_count();
      
    }
+
+
    else {
       ast_log(LOG_ERROR, "No call to hangup\n" );
       return -1;
    }
-
+   
    if(gH323Debug)
       ast_verbose("+++   ooh323_hangup\n");
 
@@ -905,7 +998,7 @@ static int ooh323_write(struct ast_channel *ast, struct ast_frame *f)
    return res;
 }
 
-static int ooh323_indicate(struct ast_channel *ast, int condition)
+static int ooh323_indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen)
 {
 
    struct ooh323_pvt *p = (struct ooh323_pvt *) ast->tech_pvt;
@@ -943,9 +1036,16 @@ static int ooh323_indicate(struct ast_channel *ast, int condition)
             ast_set_flag(p, H323_ALREADYGONE);
          }
          break;
+      case AST_CONTROL_HOLD:
+         ast_moh_start(ast, data, NULL);      
+         break;
+      case AST_CONTROL_UNHOLD:
+	     ast_moh_stop(ast);
+         break;
       case AST_CONTROL_PROCEEDING:
       case AST_CONTROL_RINGING:
       case AST_CONTROL_PROGRESS:
+      case -1:
          break;
       default:
          ast_log(LOG_WARNING,"Don't know how to indicate condition %d on %s\n",
@@ -962,10 +1062,31 @@ static int ooh323_indicate(struct ast_channel *ast, int condition)
 static int ooh323_fixup
    (struct ast_channel *oldchan, struct ast_channel *newchan)
 {
-   if(gH323Debug)
-      ast_verbose("ooh323c ooh323_fixup\n");
+   struct ooh323_pvt *p = newchan->tech_pvt;
 
-  return 0;
+   if(gH323Debug)
+      ast_verbose("--- ooh323c ooh323_fixup\n");
+
+   ast_mutex_lock(&p->lock);
+   if (p->owner != oldchan) {
+      ast_log(LOG_WARNING, "Old channel wasn't %p but was %p\n", oldchan, p->owner);
+      ast_mutex_unlock(&p->lock);
+      return -1;
+   }
+
+   if (p->owner == oldchan) {
+      p->owner = newchan;
+   }
+   else {
+      p->owner = oldchan;
+   }
+
+   ast_mutex_unlock(&p->lock);
+
+   if(gH323Debug)
+      ast_verbose("+++ ooh323c ooh323_fixup \n");
+
+   return 0;
 }
 
 
@@ -1216,7 +1337,7 @@ int ooh323_onReceivedSetup(ooCallData *call, Q931Message *pmsg)
    if(p->callerid_name)
    {
       p->username = strdup(p->callerid_name);
-      user = find_user(p->username);
+      user = find_user(p->username, call->remoteIP);
       if(user)
       {
          ast_mutex_lock(&user->lock);
@@ -1557,6 +1678,10 @@ static struct ooh323_user *build_user(const char *name, struct ast_variable *v)
          else if (!strcasecmp(v->name, "amaflags")) {
             user->amaflags = ast_cdr_amaflags2int(v->value);
          }
+         else if (!strcasecmp(v->name, "ip")) {
+            strncpy(user->mIP, v->value, sizeof(user->mIP)-1);
+            user->mUseIP = 1;
+         }
          else if (!strcasecmp(v->name, "dtmfmode")) {
             if(!strcasecmp(v->value, "rfc2833"))
                user->dtmfmode = H323_DTMF_RFC2833;
@@ -1577,7 +1702,7 @@ static struct ooh323_user *build_user(const char *name, struct ast_variable *v)
    return user;
 }
 
-static struct ooh323_peer *build_peer(const char *name, struct ast_variable *v)
+static struct ooh323_peer *build_peer(const char *name, struct ast_variable *v, int friend_type)
 {
    struct ooh323_peer *peer=NULL;
 
@@ -1595,6 +1720,9 @@ static struct ooh323_peer *build_peer(const char *name, struct ast_variable *v)
       strncpy(peer->accountcode, gAccountcode, sizeof(peer->accountcode)-1);
       peer->amaflags = gAMAFLAGS;
       peer->dtmfmode = gDTMFMode;
+      if(0 == friend_type) {
+         peer->mFriend = 1;
+      }
 
       while(v) {
          if (!strcasecmp(v->name, "h323id")) {
@@ -1731,7 +1859,7 @@ static int ooh323_reload(int fd, int argc, char *argv[])
    return 0;
 }
 
-int reload(void)
+static int reload(void *mod)
 {
    return ooh323_reload(0, 0, NULL);
 }
@@ -1744,7 +1872,8 @@ int reload_config(void)
    struct ast_variable *v;
    struct ooh323_user *user = NULL;
    struct ooh323_peer *peer = NULL;
-   char *cat, *utype;
+   char *cat;
+   const char *utype;
 
    if(gH323Debug)
       ast_verbose("---   reload_config\n");
@@ -1774,6 +1903,7 @@ int reload_config(void)
    gTOS = 0;
    strcpy(gContext, DEFAULT_CONTEXT);
    gAliasList = NULL;
+   gMediaWaitForConnect = 0;
 
    v = ast_variable_browse(cfg, "general");
    while(v) {
@@ -1789,9 +1919,24 @@ int reload_config(void)
       } 
       else if (!strcasecmp(v->name, "faststart")) {
          gFastStart = ast_true(v->value);
+         if(gFastStart)
+            ooH323EpEnableFastStart();
+         else
+            ooH323EpDisableFastStart();
+      } 
+      else if (!strcasecmp(v->name, "mediawaitforconnect")) {
+         gMediaWaitForConnect = ast_true(v->value);
+         if(gMediaWaitForConnect)
+            ooH323EpEnableMediaWaitForConnect();
+         else 
+            ooH323EpDisableMediaWaitForConnect();
       } 
       else if (!strcasecmp(v->name, "h245tunneling")) {
          gTunneling = ast_true(v->value);
+         if(gTunneling)
+            ooH323EpEnableH245Tunneling();
+         else
+            ooH323EpDisableH245Tunneling();
       }
       else if (!strcasecmp(v->name, "h323id")) {
          pNewAlias = malloc(sizeof(struct ooAliases));
@@ -1851,7 +1996,7 @@ int reload_config(void)
          }
       }
       else if (!strcasecmp(v->name, "logfile")) {
-         strncpy(gLogFile, v->name, sizeof(gLogFile)-1);
+         strncpy(gLogFile, v->value, sizeof(gLogFile)-1);
       }
       else if (!strcasecmp(v->name, "context")) {
          strncpy(gContext, v->value, sizeof(gContext)-1);
@@ -1918,10 +2063,12 @@ int reload_config(void)
    {
       if(strcasecmp(cat, "general")) 
       {
+         int friend_type = 0;
          utype = ast_variable_retrieve(cfg, cat, "type");
          if(utype)
          {
-            if(!strcmp(utype, "user") || !strcasecmp(utype, "friend"))
+		 friend_type = strcasecmp(utype, "friend");
+            if(!strcmp(utype, "user") || 0 == friend_type)
             {
                user = build_user(cat, ast_variable_browse(cfg, cat));
                if (user) 
@@ -1935,15 +2082,15 @@ int reload_config(void)
                   ast_log(LOG_WARNING, "Failed to build user %s\n", cat);
                }
             }
-            if(!strcasecmp(utype, "peer") || !strcasecmp(utype, "friend"))
+            if(!strcasecmp(utype, "peer") || 0 == friend_type)
             {
-               peer = build_peer(cat, ast_variable_browse(cfg, cat));
+               peer = build_peer(cat, ast_variable_browse(cfg, cat), friend_type);
                if(peer)
                {
-                  ast_mutex_lock(&userl.lock);
+                  ast_mutex_lock(&peerl.lock);
                   peer->next = peerl.peers;
                   peerl.peers = peer;
-                  ast_mutex_unlock(&userl.lock);
+                  ast_mutex_unlock(&peerl.lock);
                }
                else {
                   ast_log(LOG_WARNING, "Failed to build peer %s\n", cat);
@@ -2003,8 +2150,9 @@ static int ooh323_show_peer(int fd, int argc, char *argv[])
    {
       sprintf(ip_port, "%s:%d", peer->ip, peer->port);
       ast_cli(fd, "%-15.15s%s\n", "Name: ", peer->name);
-      ast_codec_pref_string(&peer->prefs, formats, 512);
-      ast_cli(fd, "%-15.15s%s\n", "Format Prefs: ", formats);
+      ast_cli(fd, "%-15.15s%s", "Format Prefs: ", "(");
+      print_codec_to_cli(fd, &peer->prefs);
+      ast_cli(fd, ")\n");
       ast_cli(fd, "%-15.15s", "DTMF Mode: ");
       if(peer->dtmfmode & H323_DTMF_RFC2833)
          ast_cli(fd, "%s\n", "rfc2833");
@@ -2065,6 +2213,24 @@ static int ooh323_show_peers(int fd, int argc, char *argv[])
    return RESULT_SUCCESS;
 }
 
+/*! \brief Print codec list from preference to CLI/manager */
+static void print_codec_to_cli(int fd, struct ast_codec_pref *pref)
+{
+	int x, codec;
+
+	for(x = 0; x < 32 ; x++) {
+		codec = ast_codec_pref_index(pref, x);
+		if (!codec)
+			break;
+		ast_cli(fd, "%s", ast_getformatname(codec));
+		ast_cli(fd, ":%d", pref->framing[x]);
+		if (x < 31 && ast_codec_pref_index(pref, x + 1))
+			ast_cli(fd, ",");
+		}
+	if (!x)
+		ast_cli(fd, "none");
+}
+
 static int ooh323_show_user(int fd, int argc, char *argv[])
 {
    struct ooh323_user *prev=NULL, *user=NULL;
@@ -2092,8 +2258,9 @@ static int ooh323_show_user(int fd, int argc, char *argv[])
    if(user)
    {
       ast_cli(fd, "%-15.15s%s\n", "Name: ", user->name);
-      ast_codec_pref_string(&user->prefs, formats, 512);
-      ast_cli(fd, "%-15.15s%s\n", "Format Prefs: ",formats);
+      ast_cli(fd, "%-15.15s%s", "Format Prefs: ", "(");
+      print_codec_to_cli(fd, &user->prefs);
+      ast_cli(fd, ")\n");
       ast_cli(fd, "%-15.15s", "DTMF Mode: ");
       if(user->dtmfmode & H323_DTMF_RFC2833)
          ast_cli(fd, "%s\n", "rfc2833");
@@ -2184,10 +2351,24 @@ static int ooh323_show_config(int fd, int argc, char *argv[])
 
    ast_cli(fd, "\nObjective Open H.323 Channel Driver's Config:\n");
    sprintf(value, "%s:%d", gIP, gPort);
-   ast_cli(fd, "%-15s%s\n", "IP:Port: ", value);
-   ast_cli(fd, "%-15s%s\n", "FastStart", gFastStart?"yes":"no");
-   ast_cli(fd, "%-15s%s\n", "Tunneling", gTunneling?"yes":"no");
-   ast_cli(fd, "%-15s%s\n", "CallerId", gCallerID);
+   ast_cli(fd, "%-20s%s\n", "IP:Port: ", value);
+   ast_cli(fd, "%-20s%s\n", "FastStart", gFastStart?"yes":"no");
+   ast_cli(fd, "%-20s%s\n", "Tunneling", gTunneling?"yes":"no");
+   ast_cli(fd, "%-20s%s\n", "CallerId", gCallerID);
+   ast_cli(fd, "%-20s%s\n", "MediaWaitForConnect", 
+      gMediaWaitForConnect?"yes":"no");
+   ast_cli(fd, "%-15s%s\n", "MediaWaitForConnect", 
+      gMediaWaitForConnect?"yes":"no");
+
+#if (0)
+   extern OOH323EndPoint gH323ep;
+   ast_cli(fd, "%-20s%s\n", "FASTSTART", 
+      (OO_TESTFLAG(gH323ep.flags, OO_M_FASTSTART) != 0) ? "yes" : "no");
+   ast_cli(fd, "%-20s%s\n", "TUNNELING", 
+      (OO_TESTFLAG(gH323ep.flags, OO_M_TUNNELING) != 0) ? "yes" : "no");
+   ast_cli(fd, "%-20s%s\n", "MEDIAWAITFORCONN",
+      (OO_TESTFLAG(gH323ep.flags, OO_M_MEDIAWAITFORCONN) != 0) ? "yes" : "no");
+#endif
 
    if(gRasGkMode == RasNoGatekeeper)
       sprintf(value, "%s", "No Gatekeeper");
@@ -2196,16 +2377,16 @@ static int ooh323_show_config(int fd, int argc, char *argv[])
    else
       sprintf(value, "%s", gGatekeeper);
 
-   ast_cli(fd,  "%-15s%s\n", "Gatekeeper:", value);
+   ast_cli(fd,  "%-20s%s\n", "Gatekeeper:", value);
 
-   ast_cli(fd,  "%-15s%s\n", "H.323 LogFile:", gLogFile);   
+   ast_cli(fd,  "%-20s%s\n", "H.323 LogFile:", gLogFile);   
 
-   ast_cli(fd,  "%-15s%s\n", "Context:", gContext);
+   ast_cli(fd,  "%-20s%s\n", "Context:", gContext);
    
-   ast_cli(fd,  "%-15s%s\n", "Capability:", 
+   ast_cli(fd,  "%-20s%s\n", "Capability:", 
            ast_getformatname_multiple(value,512,gCapability));
 
-   ast_cli(fd, "%-15s", "DTMF Mode: ");
+   ast_cli(fd, "%-20s", "DTMF Mode: ");
    if(gDTMFMode & H323_DTMF_RFC2833)
       ast_cli(fd, "%s\n", "rfc2833");
    else if (gDTMFMode & H323_DTMF_Q931)
@@ -2217,14 +2398,14 @@ static int ooh323_show_config(int fd, int argc, char *argv[])
    else
       ast_cli(fd, "%s\n", "unknown");
 
-   ast_cli(fd, "%-15s%s\n", "AccountCode: ", gAccountcode);
+   ast_cli(fd, "%-20s%s\n", "AccountCode: ", gAccountcode);
 
-   ast_cli(fd, "%-15s%s\n", "AMA flags: ", ast_cdr_flags2str(gAMAFLAGS));
+   ast_cli(fd, "%-20s%s\n", "AMA flags: ", ast_cdr_flags2str(gAMAFLAGS));
 
    pAlias = gAliasList;
    if(pAlias)
    {
-     ast_cli(fd, "%-15s\n", "Aliases: ");
+     ast_cli(fd, "%-20s\n", "Aliases: ");
    }
    while(pAlias)
    {
@@ -2294,7 +2475,7 @@ static struct ast_cli_entry cli_show_config =
 
 
 
-int load_module(void)
+static int load_module(void)
 {
    int res;
    struct ooAliases * pNewAlias=NULL;
@@ -2319,6 +2500,7 @@ int load_module(void)
    peerl.peers = NULL;
    ast_mutex_init(&peerl.lock);
          
+   //ast_register_atexit(&ast_ooh323c_exit);
 
    sched = sched_context_create();
    if (!sched) {
@@ -2397,8 +2579,15 @@ int load_module(void)
       ast_mutex_unlock(&peerl.lock);
    
 
+      if(gMediaWaitForConnect)
+         ooH323EpEnableMediaWaitForConnect();
+      else 
+         ooH323EpDisableMediaWaitForConnect();
+
       /* Fast start and tunneling options */
-      if(!gFastStart)
+      if(gFastStart)
+         ooH323EpEnableFastStart();
+      else
          ooH323EpDisableFastStart();
 
       if(!gTunneling)
@@ -2472,6 +2661,7 @@ static void *do_monitor(void *data)
          }
          ooh323_do_reload();
       }
+
       /* Check for interfaces needing to be killed */
       ast_mutex_lock(&iflock);
       struct ooh323_pvt *h323_next;
@@ -2488,6 +2678,7 @@ static void *do_monitor(void *data)
       }
       ast_mutex_unlock(&iflock);
       pthread_testcancel();
+
       /* Wait for sched or io */
       res = ast_sched_wait(sched);
       if ((res < 0) || (res > 1000)) {
@@ -2657,7 +2848,7 @@ int delete_users()
    return 0;
 }
   
-int unload_module()
+static int unload_module(void)
 {
    struct ooh323_pvt *p;
    struct ooAliases *cur=NULL, *prev=NULL;
@@ -2675,6 +2866,7 @@ int unload_module()
    ast_cli_unregister(&cli_no_debug);
    ast_rtp_proto_unregister(&ooh323_rtp);
    ast_channel_unregister(&ooh323_tech);
+   //ast_unregister_atexit(&ast_ooh323c_exit);
 
    if(gH323Debug) {
       ast_verbose("  unload_module - hanging up all interfaces\n");
@@ -2787,20 +2979,40 @@ int unload_module()
 
 
 
-static struct ast_rtp *ooh323_get_rtp_peer(struct ast_channel *chan)
+static enum ast_rtp_get_result ooh323_get_rtp_peer(struct ast_channel *chan, struct ast_rtp **rtp)
 {
-   struct ooh323_pvt *p;
-   p = (struct ooh323_pvt *) chan->tech_pvt;
+   struct ooh323_pvt *p = NULL;
+   enum ast_rtp_get_result res = AST_RTP_TRY_PARTIAL;
 
-   if (p && p->rtp) {
-      return p->rtp;
+   if (!(p = (struct ooh323_pvt *) chan->tech_pvt))
+	return AST_RTP_GET_FAILED;
+
+   *rtp = p->rtp;
+
+   if (!(p->rtp)) {
+	return AST_RTP_GET_FAILED;
    }
-   return NULL;
+   res = AST_RTP_TRY_NATIVE;
+
+   return res;
 }
 
-static struct ast_rtp *ooh323_get_vrtp_peer(struct ast_channel *chan)
+static enum ast_rtp_get_result ooh323_get_vrtp_peer(struct ast_channel *chan, struct ast_rtp **rtp)
 {
-   return NULL;
+   struct ooh323_pvt *p = NULL;
+   enum ast_rtp_get_result res = AST_RTP_TRY_PARTIAL;
+
+   if (!(p = (struct ooh323_pvt *) chan->tech_pvt))
+        return AST_RTP_GET_FAILED;
+
+   *rtp = p->vrtp;
+
+   if (!(p->rtp)) {
+        return AST_RTP_GET_FAILED;
+   }
+   res = AST_RTP_TRY_NATIVE;
+
+   return res;
 }
 
 
@@ -2835,6 +3047,8 @@ int ooh323_convertAsteriskCapToH323Cap(int cap)
          return OO_G729A;
       case AST_FORMAT_G723_1:
          return OO_G7231;
+      case AST_FORMAT_H263:
+         return OO_H263VIDEO;
       default:
          ast_log(LOG_NOTICE, "Don't know how to deal with mode %s\n", 
                         ast_getformatname_multiple(formats,512,cap));
@@ -2878,22 +3092,18 @@ int configure_local_rtp(struct ooh323_pvt *p, ooCallData *call)
    struct sockaddr_in us;
    ooMediaInfo mediaInfo;
    int x, format=0;     
-   static char iabuf[INET_ADDRSTRLEN];
 
    if(gH323Debug)
       ast_verbose("---   configure_local_rtp\n");
 
+   if (p->rtp){
+	ast_rtp_codec_setpref(p->rtp, &p->prefs);
+   }
 
    /* figure out our local RTP port and tell the H.323 stack about it*/
    ast_rtp_get_us(p->rtp, &us);
 
-   if(!ast_inet_ntoa(iabuf, sizeof(iabuf), us.sin_addr))
-   {
-      ast_log(LOG_ERROR, "Unable to allocate rtp_info, this is very bad.\n");
-      ast_mutex_unlock(&p->lock);
-      return 0;
-   }
-   strncpy(mediaInfo.lMediaIP, iabuf, sizeof(mediaInfo.lMediaIP)-1);
+   strncpy(mediaInfo.lMediaIP, ast_inet_ntoa(us.sin_addr), sizeof(mediaInfo.lMediaIP)-1);
    mediaInfo.lMediaPort = ntohs(us.sin_port);
    mediaInfo.lMediaCntrlPort = mediaInfo.lMediaPort +1;
    for(x=0; 0 != (format=ast_codec_pref_index(&p->prefs, x)); x++)
@@ -3104,7 +3314,7 @@ int ooh323_convert_hangupcause_h323ToAsterisk(int cause)
       case OO_REASON_REMOTE_CONGESTED:
       case OO_REASON_LOCAL_CONGESTED:
          return AST_CAUSE_CONGESTION;
-      case OO_REASON_REMOTE_NOANSWER:     
+      case OO_REASON_REMOTE_NOANSWER:
          return AST_CAUSE_NO_ANSWER;
       case OO_REASON_UNKNOWN: 
       case OO_REASON_INVALIDMESSAGE:
@@ -3119,23 +3329,9 @@ int ooh323_convert_hangupcause_h323ToAsterisk(int cause)
    return 0;
 }
 
-int usecount()
+void ast_ooh323c_exit()
 {
-   int res;
-   ast_mutex_lock(&usecnt_lock);
-   res = usecnt;
-   ast_mutex_unlock(&usecnt_lock);
-   return res;
+   ooGkClientDestroy();
 }
 
-char *key()
-{
-   return ASTERISK_GPL_KEY;
-}
-
-char *description()
-{
-   return (char*)desc;
-}
-
-
+AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Objective Systems H323 Channel");
