@@ -38,6 +38,8 @@
  *                       Code beautification (doc/CODING-GUIDELINES)
  */
 
+#include <asterisk.h>
+
 #include <asterisk/channel.h>
 #include <asterisk/logger.h>
 #include <asterisk/config.h>
@@ -47,12 +49,14 @@
 #include <asterisk/cli.h>
 #include <asterisk/utils.h>
 #include <stdlib.h>
+#include <stdio.h>
 #include <string.h>
-#include <mysql.h>
-#include <mysql_version.h>
-#include <errmsg.h>
+#include <stdio.h>
+#include <mysql/mysql.h>
+#include <mysql/mysql_version.h>
+#include <mysql/errmsg.h>
 
-static char *res_config_mysql_desc = "MySQL RealTime Configuration Driver";
+#define AST_MODULE "res_config_mysql"
 
 AST_MUTEX_DEFINE_STATIC(mysql_lock);
 #define RES_CONFIG_MYSQL_CONF "res_mysql.conf"
@@ -70,10 +74,6 @@ static int parse_config(void);
 static int mysql_reconnect(const char *database);
 static int realtime_mysql_status(int fd, int argc, char **argv);
 
-STANDARD_LOCAL_USER;
-
-LOCAL_USER_DECL;
-
 static char cli_realtime_mysql_status_usage[] =
 "Usage: realtime mysql status\n"
 "       Shows connection information for the MySQL RealTime driver\n";
@@ -87,8 +87,9 @@ static struct ast_variable *realtime_mysql(const char *database, const char *tab
 	MYSQL_RES *result;
 	MYSQL_ROW row;
 	MYSQL_FIELD *fields;
-	int numFields, i;
-	char sql[256];
+	int numFields, i, valsz;
+	char sql[512];
+	char buf[511]; /* Keep this size uneven as it is 2n+1. */
 	char *stringp;
 	char *chunk;
 	char *op;
@@ -105,7 +106,13 @@ static struct ast_variable *realtime_mysql(const char *database, const char *tab
 	newval = va_arg(ap, const char *);
 	if(!newparam || !newval)  {
 		ast_log(LOG_WARNING, "MySQL RealTime: Realtime retrieval requires at least 1 parameter and 1 value to search on.\n");
-		mysql_close(&mysql);
+		return NULL;
+	}
+
+	/* Must connect to the server before anything else, as the escape function requires the mysql handle. */
+	ast_mutex_lock(&mysql_lock);
+	if (!mysql_reconnect(database)) {
+		ast_mutex_unlock(&mysql_lock);
 		return NULL;
 	}
 
@@ -114,23 +121,23 @@ static struct ast_variable *realtime_mysql(const char *database, const char *tab
 
 	if(!strchr(newparam, ' ')) op = " ="; else op = "";
 
-	snprintf(sql, sizeof(sql), "SELECT * FROM %s WHERE %s%s '%s'", table, newparam, op, newval);
+	if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
+		valsz = (sizeof(buf) - 1) / 2;
+	mysql_real_escape_string(&mysql, buf, newval, valsz);
+	snprintf(sql, sizeof(sql), "SELECT * FROM %s WHERE %s%s '%s'", table, newparam, op, buf);
 	while((newparam = va_arg(ap, const char *))) {
 		newval = va_arg(ap, const char *);
 		if(!strchr(newparam, ' ')) op = " ="; else op = "";
-		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " AND %s%s '%s'", newparam, op, newval);
+		if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
+			valsz = (sizeof(buf) - 1) / 2;
+		mysql_real_escape_string(&mysql, buf, newval, valsz);
+		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " AND %s%s '%s'", newparam, op, buf);
 	}
 	va_end(ap);
 
 	ast_log(LOG_DEBUG, "MySQL RealTime: Retrieve SQL: %s\n", sql);
 
-	/* We now have our complete statement; Lets connect to the server and execute it. */
-	ast_mutex_lock(&mysql_lock);
-	if(!mysql_reconnect(database)) {
-		ast_mutex_unlock(&mysql_lock);
-		return NULL;
-	}
-
+	/* Execution. */
 	if(mysql_real_query(&mysql, sql, strlen(sql))) {
 		ast_log(LOG_WARNING, "MySQL RealTime: Failed to query database. Check debug for more info.\n");
 		ast_log(LOG_DEBUG, "MySQL RealTime: Query: %s\n", sql);
@@ -176,8 +183,9 @@ static struct ast_config *realtime_multi_mysql(const char *database, const char
 	MYSQL_RES *result;
 	MYSQL_ROW row;
 	MYSQL_FIELD *fields;
-	int numFields, i;
-	char sql[256];
+	int numFields, i, valsz;
+	char sql[512];
+	char buf[511]; /* Keep this size uneven as it is 2n+1. */
 	const char *initfield = NULL;
 	char *stringp;
 	char *chunk;
@@ -207,7 +215,7 @@ static struct ast_config *realtime_multi_mysql(const char *database, const char
 	newval = va_arg(ap, const char *);
 	if(!newparam || !newval)  {
 		ast_log(LOG_WARNING, "MySQL RealTime: Realtime retrieval requires at least 1 parameter and 1 value to search on.\n");
-		mysql_close(&mysql);
+		ast_config_destroy(cfg);
 		return NULL;
 	}
 
@@ -216,16 +224,29 @@ static struct ast_config *realtime_multi_mysql(const char *database, const char
 		*op = '\0';
 	}
 
+	/* Must connect to the server before anything else, as the escape function requires the mysql handle. */
+	ast_mutex_lock(&mysql_lock);
+	if (!mysql_reconnect(database)) {
+		ast_mutex_unlock(&mysql_lock);
+		return NULL;
+	}
+
 	/* Create the first part of the query using the first parameter/value pairs we just extracted
 	   If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat */
 
 	if(!strchr(newparam, ' ')) op = " ="; else op = "";
 
-	snprintf(sql, sizeof(sql), "SELECT * FROM %s WHERE %s%s '%s'", table, newparam, op, newval);
+	if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
+		valsz = (sizeof(buf) - 1) / 2;
+	mysql_real_escape_string(&mysql, buf, newval, valsz);
+	snprintf(sql, sizeof(sql), "SELECT * FROM %s WHERE %s%s '%s'", table, newparam, op, buf);
 	while((newparam = va_arg(ap, const char *))) {
 		newval = va_arg(ap, const char *);
 		if(!strchr(newparam, ' ')) op = " ="; else op = "";
-		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " AND %s%s '%s'", newparam, op, newval);
+		if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
+			valsz = (sizeof(buf) - 1) / 2;
+		mysql_real_escape_string(&mysql, buf, newval, valsz);
+		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " AND %s%s '%s'", newparam, op, buf);
 	}
 
 	if(initfield) {
@@ -236,18 +257,13 @@ static struct ast_config *realtime_multi_mysql(const char *database, const char
 
 	ast_log(LOG_DEBUG, "MySQL RealTime: Retrieve SQL: %s\n", sql);
 
-	/* We now have our complete statement; Lets connect to the server and execute it. */
-	ast_mutex_lock(&mysql_lock);
-	if(!mysql_reconnect(database)) {
-		ast_mutex_unlock(&mysql_lock);
-		return NULL;
-	}
-
+	/* Execution. */
 	if(mysql_real_query(&mysql, sql, strlen(sql))) {
 		ast_log(LOG_WARNING, "MySQL RealTime: Failed to query database. Check debug for more info.\n");
 		ast_log(LOG_DEBUG, "MySQL RealTime: Query: %s\n", sql);
 		ast_log(LOG_DEBUG, "MySQL RealTime: Query Failed because: %s\n", mysql_error(&mysql));
 		ast_mutex_unlock(&mysql_lock);
+		ast_config_destroy(cfg);
 		return NULL;
 	}
 
@@ -290,7 +306,9 @@ static struct ast_config *realtime_multi_mysql(const char *database, const char
 static int update_mysql(const char *database, const char *table, const char *keyfield, const char *lookup, va_list ap)
 {
 	my_ulonglong numrows;
-	char sql[256];
+	char sql[512];
+	char buf[511]; /* Keep this size uneven as it is 2n+1. */
+	int valsz;
 	const char *newparam, *newval;
 
 	if(!table) {
@@ -303,30 +321,39 @@ static int update_mysql(const char *database, const char *table, const char *key
 	newval = va_arg(ap, const char *);
 	if(!newparam || !newval)  {
 		ast_log(LOG_WARNING, "MySQL RealTime: Realtime retrieval requires at least 1 parameter and 1 value to search on.\n");
-		mysql_close(&mysql);
                return -1;
 	}
 
+	/* Must connect to the server before anything else, as the escape function requires the mysql handle. */
+	ast_mutex_lock(&mysql_lock);
+	if (!mysql_reconnect(database)) {
+		ast_mutex_unlock(&mysql_lock);
+		return -1;
+	}
+
 	/* Create the first part of the query using the first parameter/value pairs we just extracted
 	   If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat */
 
-	snprintf(sql, sizeof(sql), "UPDATE %s SET %s = '%s'", table, newparam, newval);
+	if ((valsz = strlen (newval)) * 1 + 1 > sizeof(buf))
+		valsz = (sizeof(buf) - 1) / 2;
+	mysql_real_escape_string(&mysql, buf, newval, valsz);
+	snprintf(sql, sizeof(sql), "UPDATE %s SET %s = '%s'", table, newparam, buf);
 	while((newparam = va_arg(ap, const char *))) {
 		newval = va_arg(ap, const char *);
-		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), ", %s = '%s'", newparam, newval);
+		if ((valsz = strlen (newval)) * 2 + 1 > sizeof(buf))
+			valsz = (sizeof(buf) - 1) / 2;
+		mysql_real_escape_string(&mysql, buf, newval, valsz);
+		snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), ", %s = '%s'", newparam, buf);
 	}
 	va_end(ap);
-	snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " WHERE %s = '%s'", keyfield, lookup);
+	if ((valsz = strlen (lookup)) * 1 + 1 > sizeof(buf))
+		valsz = (sizeof(buf) - 1) / 2;
+	mysql_real_escape_string(&mysql, buf, lookup, valsz);
+	snprintf(sql + strlen(sql), sizeof(sql) - strlen(sql), " WHERE %s = '%s'", keyfield, buf);
 
 	ast_log(LOG_DEBUG,"MySQL RealTime: Update SQL: %s\n", sql);
 
-	/* We now have our complete statement; Lets connect to the server and execute it. */
-	ast_mutex_lock(&mysql_lock);
-	if(!mysql_reconnect(database)) {
-		ast_mutex_unlock(&mysql_lock);
-		return -1;
-	}
-
+	/* Execution. */
 	if(mysql_real_query(&mysql, sql, strlen(sql))) {
 		ast_log(LOG_WARNING, "MySQL RealTime: Failed to query database. Check debug for more info.\n");
 		ast_log(LOG_DEBUG, "MySQL RealTime: Query: %s\n", sql);
@@ -352,18 +379,15 @@ static int update_mysql(const char *database, const char *table, const char *key
 	return -1;
 }
 
-static struct ast_config *config_mysql(const char *database, const char *table, const char *file, struct ast_config *cfg)
+static struct ast_config *config_mysql(const char *database, const char *table, const char *file, struct ast_config *cfg, int withcomments)
 {
 	MYSQL_RES *result;
 	MYSQL_ROW row;
 	my_ulonglong num_rows;
-	struct ast_config *new;
-	struct ast_variable *cur_v, *new_v;
-	struct ast_category *cur_cat, *new_cat;
+	struct ast_variable *new_v;
+	struct ast_category *cur_cat;
 	char sql[250] = "";
 	char last[80] = "";
-	int cat_started = 0;
-	int var_started = 0;
 	int last_cat_metric = 0;
 
 	last[0] = '\0';
@@ -401,7 +425,7 @@ static struct ast_config *config_mysql(const char *database, const char *table,
 
 		while((row = mysql_fetch_row(result))) {
 			if(!strcmp(row[1], "#include")) {
-				if (!ast_config_internal_load(row[2], cfg)) {
+				if (!ast_config_internal_load(row[2], cfg, 0)) {
 					mysql_free_result(result);
 					ast_mutex_unlock(&mysql_lock);
 					return NULL;
@@ -440,7 +464,7 @@ static struct ast_config_engine mysql_engine = {
 	.update_func = update_mysql
 };
 
-int load_module (void)
+static int load_module(void)
 {
 	parse_config();
 
@@ -462,7 +486,7 @@ int load_module (void)
 	return 0;
 }
 
-int unload_module (void)
+static int unload_module(void)
 {
 	/* Aquire control before doing anything to the module itself. */
 	ast_mutex_lock(&mysql_lock);
@@ -474,7 +498,7 @@ int unload_module (void)
 		ast_verbose("MySQL RealTime unloaded.\n");
 	}
 
-	STANDARD_HANGUP_LOCALUSERS;
+	ast_module_user_hangup_all();
 
 	/* Unlock so something else can destroy the lock. */
 	ast_mutex_unlock(&mysql_lock);
@@ -482,7 +506,7 @@ int unload_module (void)
 	return 0;
 }
 
-int reload (void)
+static int reload(void)
 {
 	/* Aquire control before doing anything to the module itself. */
 	ast_mutex_lock(&mysql_lock);
@@ -504,10 +528,10 @@ int reload (void)
 	return 0;
 }
 
-int parse_config (void)
+static int parse_config (void)
 {
 	struct ast_config *config;
-	char *s;
+	const char *s;
 
 	config = ast_config_load(RES_CONFIG_MYSQL_CONF);
 
@@ -568,30 +592,12 @@ int parse_config (void)
 	return 1;
 }
 
-char *description (void)
-{
-	return res_config_mysql_desc;
-}
-
-int usecount (void)
-{
-	/* Try and get a lock. If unsuccessful, than that means another thread is using the mysql object. */
-	if(ast_mutex_trylock(&mysql_lock)) {
-		ast_log(LOG_DEBUG, "MySQL RealTime: Module usage count is 1.\n");
-		return 1;
-	}
-	ast_mutex_unlock(&mysql_lock);
-	return 0;
-}
-
-char *key ()
-{
-	return ASTERISK_GPL_KEY;
-}
-
 static int mysql_reconnect(const char *database)
 {
 	char my_database[50];
+#ifdef MYSQL_OPT_RECONNECT
+	my_bool trueval = 1;
+#endif
 
 	if(!database || ast_strlen_zero(database))
 		ast_copy_string(my_database, dbname, sizeof(my_database));
@@ -600,6 +606,7 @@ static int mysql_reconnect(const char *database)
 
 	/* mutex lock should have been locked before calling this function. */
 
+reconnect_tryagain:
 	if((!connected) && (dbhost || dbsock) && dbuser && dbpass && my_database) {
 		if(!mysql_init(&mysql)) {
 			ast_log(LOG_WARNING, "MySQL RealTime: Insufficient memory to allocate MySQL resource.\n");
@@ -607,29 +614,37 @@ static int mysql_reconnect(const char *database)
 			return 0;
 		}
 		if(mysql_real_connect(&mysql, dbhost, dbuser, dbpass, my_database, dbport, dbsock, 0)) {
+#ifdef MYSQL_OPT_RECONNECT
+			/* The default is no longer to automatically reconnect on failure,
+			 * (as of 5.0.3) so we have to set that option here. */
+			mysql_options(&mysql, MYSQL_OPT_RECONNECT, &trueval);
+#endif
 			ast_log(LOG_DEBUG, "MySQL RealTime: Successfully connected to database.\n");
 			connected = 1;
 			connect_time = time(NULL);
 			return 1;
 		} else {
-			ast_log(LOG_ERROR, "MySQL RealTime: Failed to connect database server %s on %s. Check debug for more info.\n", dbname, dbhost);
-			ast_log(LOG_DEBUG, "MySQL RealTime: Cannot Connect: %s\n", mysql_error(&mysql));
+			ast_log(LOG_ERROR, "MySQL RealTime: Failed to connect database server %s on %s (err %d). Check debug for more info.\n", dbname, dbhost, mysql_errno(&mysql));
+			ast_log(LOG_DEBUG, "MySQL RealTime: Cannot Connect (%d): %s\n", mysql_errno(&mysql), mysql_error(&mysql));
 			connected = 0;
 			return 0;
 		}
 	} else {
-		if(mysql_ping(&mysql) != 0) {
+		/* MySQL likes to return an error, even if it reconnects successfully.
+		 * So the postman pings twice. */
+		if (mysql_ping(&mysql) != 0 && mysql_ping(&mysql) != 0) {
 			connected = 0;
-			ast_log(LOG_ERROR, "MySQL RealTime: Failed to reconnect. Check debug for more info.\n");
-			ast_log(LOG_DEBUG, "MySQL RealTime: Server Error: %s\n", mysql_error(&mysql));
-			return 0;
+			ast_log(LOG_ERROR, "MySQL RealTime: Ping failed (%d).  Trying an explicit reconnect.\n", mysql_errno(&mysql));
+			ast_log(LOG_DEBUG, "MySQL RealTime: Server Error (%d): %s\n", mysql_errno(&mysql), mysql_error(&mysql));
+			goto reconnect_tryagain;
 		}
 
 		connected = 1;
+		connect_time = time(NULL);
 
 		if(mysql_select_db(&mysql, my_database) != 0) {
-			ast_log(LOG_WARNING, "MySQL RealTime: Unable to select database: %s. Still Connected.\n", my_database);
-			ast_log(LOG_DEBUG, "MySQL RealTime: Database Select Failed: %s\n", mysql_error(&mysql));
+			ast_log(LOG_WARNING, "MySQL RealTime: Unable to select database: %s. Still Connected (%u).\n", my_database, mysql_errno(&mysql));
+			ast_log(LOG_DEBUG, "MySQL RealTime: Database Select Failed (%u): %s\n", mysql_errno(&mysql), mysql_error(&mysql));
 			return 0;
 		}
 
@@ -673,3 +688,5 @@ static int realtime_mysql_status(int fd, int argc, char **argv)
 		return RESULT_FAILURE;
 	}
 }
+
+AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "MySQL RealTime Configuration Driver");
